/* DEC Rainbow 100 Game Input/Output Routines
 *
 * Based on JZip's original mscio.c routines
 *
 * Copyright (c) 2011 Jeffrey Armstrong <jeff@rainbow-100.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <dos.h>
#include <sys/timeb.h>
#include <sys/types.h>

#include "ztypes.h"
#include "rbscr.h"


/* The Rainbow can have different screen widths */
#define SCREEN_WIDTH    80
#define SCREEN_HEIGHT   24

/* ----------- Rainbow Functions ---------------- */

int _rb_r, _rb_c, _rb_top, _rb_bottom;
int _rb_status_select, _rb_status_size;
int _rb_cursorenable;
int _rb_wait_on;

/* Retrieves Level 1 console input (16bit).
 * Note that this function blocks until input
 * is received.
 */
int inkey()
{
union REGS r;

        do {
                r.x.di = (unsigned int)0x06;
                int86(0x18,&r,&r);
        } while(r.h.cl != 0xFF);
        
        return (int)r.x.ax;
}

/* Non-blocking key input */
int getkey()
{
union REGS r;

        r.x.di = (unsigned int)0x06;
        int86(0x18,&r,&r);
        if(r.h.cl != 0xFF)
            return 0;
        else
            return (int)r.x.ax;
}
char rgetchar()
{
union REGS r;

        r.x.di = (unsigned int)0x02;
        int86(0x18,&r,&r);
        if(r.h.cl != 0xFF)
            return 0;
        else
            return (char)r.h.al;
}

/* Disables the cursor during the game */
void rdisablecursor()
{
union REGS r;

        r.x.di = (unsigned int)0x08;
        int86(0x18,&r,&r);
        _rb_cursorenable = 0;
}

/* Enables the cursor when the game is over */
void renablecursor()
{
union REGS r;

        r.x.di = (unsigned int)0x0A;
        int86(0x18,&r,&r);
        _rb_cursorenable = 1;
        
}

/* Puts a single character at a given position */
void rputchar(char c, int row, int col)
{
        rbs_move(row-1,col-1);
        rbs_putc(c);
}

/* Puts a string at the give position.  No width calculations
 * are made, so beware exceeding 132 columns.
 */
void rputstr(char *string, int row, int col)
{
int i;
 
        for(i=0;string[i] != '\0'; i++) {
                rputchar(string[i],row,col+i);
        }
}

/* Enable/disable wait */
void set_wait(int i)
{
union REGS r;

        r.h.al = (char)1;
        if(i == 1)
                r.x.di = (unsigned int)0x10;
        else
                r.x.di = (unsigned int)0x12;
        int86(0x18,&r,&r);
}

/* Ring the bell (except on 100A) */
void rbell()
{
union REGS r;

        r.x.di = (unsigned int)0x1E;
        int86(0x18,&r,&r);
}


/* Clears a single row on the screen by writing spaces using
 * fast video routines.
 */
void rclearrow(int row)
{
        rbs_clearrow(row);
}

void clrscr()
{
        rbs_clearscreen();
}

/* #define _outtext(str)       rputstr(str, _rb_r, _rb_c); _rb_c += strlen(str) */

void _outtext(char *str)
{
                
        rputstr(str, _rb_r, _rb_c); 
        _rb_c += strlen(str);
        
        rbs_setcursor(_rb_r,_rb_c,0);
}

/* Pass-through color functions */
#define _settextcolor(x)    x
#define _setbkcolor(x)      x
#define _gettextcolor(  )   1
#define _getbkcolor(  )     0

/* Pass-through clear function - needs implementation */
/* #define _clearscreen(x)     x */
void _clearscreen(int x)
{
        rbs_clearscreen();
}

#define _GWINDOW            1
#define _GSCROLLUP          1
#define _GCURSOROFF         0
#define _GCURSORON          1

void _displaycursor(int action)
{
    if(action == _GCURSOROFF)
        rdisablecursor();
    else
        renablecursor();
}

void _gettextwindow(short int *top, short int *left, short int *bottom, short int *right)
{
    *top = (short int)_rb_top;
    *bottom = (short int)_rb_bottom;
    *left = (short int)0;
    *right = (short int)SCREEN_WIDTH;
}

void _settextwindow(short int top, short int left, short int bottom, short int right)
{
    _rb_top = (int)top;
    _rb_bottom = (int)bottom;
}

void _settextposition(x,y)
{
        _rb_r = x > 0 ? x : 1;
        _rb_c = y;
        
        rbs_move(x,y);
        rbs_setcursor(_rb_r,_rb_c,0);
}

void _scrolltextwindow(int ignored)
{ 
int r,c;

    r = _rb_r; c = _rb_c;
    
    _settextposition(_rb_bottom,SCREEN_WIDTH-1);
    rbs_scrollup();    
    r--;

    _rb_r = r; _rb_c = c;
}

/* End Rainbow emulation functions */

#define                 BLACK      0
#define                 BLUE       1
#define                 UNDERSCORE 1
#define                 GREEN      2
#define                 CYAN       3
#define                 RED        4
#define                 MAGENTA    5
#define                 BROWN      6
#define                 WHITE      7

#define                 BRIGHT     0x08
#define                 FLASH      0x10

#define                 PALETTE_NIL             -1
#define                 PALETTE_COLOUR          0
#define                 PALETTE_BLACK_AND_WHITE 1
#define                 PALETTE_MONOCHROME      2
int iPalette = PALETTE_NIL;

static int screen_started = FALSE;
static int cursor_saved = OFF;
static int saved_row = 0;
static int saved_col = 0;
static int current_fg = WHITE;
static int current_bg = BLACK;
static int fgBright = 0;


#if defined JZIPVER
int BUFFER_SIZE;
char *commands;
static int space_avail;
static int ptr1, ptr2 = 0;
static int end_ptr = 0;
static int row, head_col;

static int init_char_pos, curr_char_pos;
static int loop, tail_col;
#endif


void initialize_screen( void )
{
    set_wait(1);

    screen_rows = SCREEN_HEIGHT;
    screen_cols = SCREEN_WIDTH;

    _rb_r = 1;
    _rb_c = 1;

    _rb_status_select = 0;
    _rb_status_size = 0;
    
    _rb_top = 1;
    _rb_bottom = SCREEN_HEIGHT;

    _rb_cursorenable=0;
    rdisablecursor();

    /* Disables auto-repeat */
    /* printf("%c[?8l",(char)27); */
    
    rbs_init();
   
    clrscr();
    move_cursor(1, 1);
    set_attribute(NORMAL);
    clrscr();

    /*move_cursor( SCREEN_HEIGHT / 2 - 1, ( SCREEN_WIDTH - sizeof ( JZIPVER ) ) / 2 );
     * _outtext( JZIPVER );
     * move_cursor( SCREEN_HEIGHT / 2, ( SCREEN_WIDTH - sizeof ( "The story is loading..." ) ) / 2 );
     * _outtext( "The story is loading..." );
     */

     move_cursor( SCREEN_HEIGHT / 3 - 2, ( SCREEN_WIDTH/2 - sizeof ( JZIPVER ) ) / 2 );
     printf("%c#3%s",(char)27,JZIPVER);
     move_cursor( SCREEN_HEIGHT / 3 - 1, ( SCREEN_WIDTH/2 - sizeof ( JZIPVER ) ) / 2 );
     printf("%c#4%s",(char)27,JZIPVER);
     
     move_cursor( SCREEN_HEIGHT / 2, ( SCREEN_WIDTH/2 - sizeof ( "The story is loading..." ) ) / 2 );
     printf("%c#6The story is loading...",(char)27);

    screen_started = TRUE;

    h_interpreter = INTERP_MSDOS;
    JTERP = INTERP_MSDOS;

    _rb_wait_on = 1;

}                               /* initialize_screen */


void restart_screen( void )
{

   zbyte_t config_byte = get_byte( H_CONFIG );

   cursor_saved = OFF;

#if defined JZIPVER
   {
      zbyte_t high = 1, low = 0;

      set_byte( H_STANDARD_HIGH, high );
      set_byte( H_STANDARD_LOW, low );
   }
#endif

   if ( h_type == V3 )
   {
      config_byte |= CONFIG_MAX_DATA | CONFIG_WINDOWS;
      config_byte &= ~0x40 /* not variable pitch */  & ~CONFIG_NOSTATUSLINE;
   }
   else if ( h_type >= V4 )
   {
      config_byte |= 0x04 /* Bold */  | 0x80 /* Timer */ ;
      if ( iPalette == PALETTE_MONOCHROME )
      {
         config_byte |= CONFIG_EMPHASIS;
      }
      else
      {
         config_byte &= ~CONFIG_EMPHASIS;
      }                         /* endif */
      config_byte &= ~0x10 /* no switch to fixed */ ;
      if ( h_type >= V5 )
      {
         config_byte &= ~0x20 /* no sound */ ;
         if ( iPalette == PALETTE_COLOUR )
         {
            config_byte |= CONFIG_COLOUR;
         }
         else
         {
            config_byte &= ~CONFIG_COLOUR;
         }                      /* endif */
      }                         /* endif */
   }                            /* endif */

   set_byte( H_CONFIG, config_byte );

   set_word( H_FLAGS, get_word( H_FLAGS ) & ~GRAPHICS_FLAG & ~0x20 /* Mouse */
             & ~NEW_SOUND_FLAG );

}                               /* restart_screen */


void reset_screen( void )
{

   if ( screen_started == TRUE )
   {
      output_new_line(  );
      output_string( "[Hit any key to exit.]" );
      ( void ) input_character( 0 );
      output_new_line(  );

      _settextcolor( WHITE );
      _setbkcolor( BLACK );
      clear_screen(  );
      move_cursor( 1, 1 );
      
        renablecursor();
        /* Enables auto-repeat */
      /* printf("%c[?8h",(char)27); */
    
   }

   screen_started = FALSE;

}                               /* reset_screen */

void set_status_size(int s)
{
        _rb_status_size = s;
        
        rbs_setscrollregion(s,SCREEN_HEIGHT);
}

void clear_screen( void )
{

   _clearscreen( _GWINDOW );
   lines_written = 0;
   if ( h_type >= V4 )
   {
      set_status_size( 0 );
   }                            /* endif */

   if(_rb_wait_on == 1) {
        set_wait(0);
        _rb_wait_on = 0;
   }
}                               /* clear_screen */


void create_status_window( void )
{

}                               /* create_status_window */


void delete_status_window( void )
{

}                               /* delete_status_window */


void select_status_window( void )
{

    save_cursor_position(  );
    _rb_status_select = 1;
    /* rdisablecursor(); */
    move_cursor(1,1);
}                               /* select_status_window */


void select_text_window( void )
{
    _rb_status_select = 0;
    restore_cursor_position(  );
    /* renablecursor(); */
}                               /* select_text_window */


void clear_line( void )
{

   int i, row, col;

   get_cursor_position( &row, &col );
   rbs_clearrow(row);
   move_cursor( row, col );

}                               /* clear_line */


void clear_text_window( void )
{
int row, col, i;

    get_cursor_position( &row, &col );
    for(i=_rb_status_size+1;i<=screen_rows;i++)
        rbs_clearrow(i);

    move_cursor( row, col );

}                               /* clear_text_window */


void clear_status_window( void )
{
int row, col, i;

    get_cursor_position( &row, &col );
    for(i=1;i<=_rb_status_size;i++)
        rclearrow(i);
        
    move_cursor( row, col );

}                               /* clear_status_window */


void move_cursor( int row, int col )
{
    if(col == 0) col++;
    
    _settextposition( row, col );
    rbs_setcursor(row,col,0);
}


void get_cursor_position(int *row, int *col)
{
     *row = _rb_r;
     *col = _rb_c;
    
    /* rbs_getcursor(row,col); */
}                               /* get_cursor_position */


void save_cursor_position( void )
{
    if ( cursor_saved == OFF ) {
        get_cursor_position( &saved_row, &saved_col );
        cursor_saved = ON;
    }
}                               /* save_cursor_position */


void restore_cursor_position( void )
{
    if ( cursor_saved == ON ) {
        move_cursor( saved_row, saved_col );
        cursor_saved = OFF;
    }
}                               /* restore_cursor_position */


void set_attribute( int attribute )
{
char attr;

    attr = RB_AT_NONE;
    
    if(attribute==NORMAL) {
        rbs_seta(RB_AT_NONE);
        return;
    }
  
    if(attribute & REVERSE)
        attr += RB_AT_REVERSE;

    if(attribute & BOLD)
        attr -= RB_AT_BOLD;
        
    if(attribute & EMPHASIS)
        attr -= RB_AT_UNDER;

    rbs_seta(attr);
}                               /* set_attribute */


void display_char( int c )
{
   switch ( c )
   {
      case '\n':
         scroll_line(  );
         break;
      case '\a':
         rbell();         
         break;
      default:
         {
         if(_rb_cursorenable == 0)
                rputchar((char)c, _rb_r, _rb_c);
         else {
                rdisablecursor();
                rputchar((char)c, _rb_r, _rb_c);
                rbs_setcursor(_rb_r,_rb_c+1,0);
                renablecursor();
         }
         _rb_c++;
        }
   }                            /* endswitch */

}                               /* display_char */


static long convert_key( long c )
{

   /* Translate extended keys to Infocom-determined single bytes above 127 */
   switch ( c )
   {
      case 'H':                /* Up arrow */
         c = 0x81;
         break;
      case 'P':                /* Down arrow */
         c = 0x82;
         break;
      case 'K':                /* Left arrow */
         c = 0x83;
         break;
      case 's':                /* Ctrl-Left arrow */
         c = 0x20000001;
         break;
      case 'M':                /* Right arrow */
         c = 0x84;
         break;
      case 't':                /* Ctrl-Right arrow */
         c = 0x20000002;
         break;
      case ';':                /* Function keys F1 to F10 */
      case '<':
      case '=':
      case '>':
      case '?':
      case '@':
      case 'A':
      case 'B':
      case 'C':
      case 'D':
         c = ( c - ';' ) + 0x85;
         break;
      case 133:                /* Function key F11 */
         c = 0x8F;
         break;
      case 134:                /* Function key F12 */
         c = 0x90;
         break;
      case 'R':                /* Ins */
         c = 0x91;
         break;
      case 'S':                /* Del */
         c = 0x20000000;
         break;
      case 'O':                /* End (SW) */
         c = 0x92;
         break;
      case 'Q':                /* PgDn (SE) */
         c = 0x94;
         break;
      case 'G':                /* Home (NW) */
         c = 0x98;
         break;
      case 'I':                /* PgUp (NE) */
         c = 0x9A;
         break;
      case 'L':                /* Unnamed (Walk Around) */
         c = 0x96;
         break;
      default:
        rbell(); 
        c = 0;
   }                            /* endswitch */
   return c;

}


static int read_key( void )
{
    return rgetchar();
}

static int timed_read_key( int target_second, int target_millisecond )
{
struct timeb timenow;
struct tm *tmptr;
int c;
union REGS r;
    
    for( ;; ) {

        do {
            ftime( &timenow );
            tmptr = gmtime( &timenow.time );
            
            r.x.di = (unsigned int)0x06;
            int86(0x18,&r,&r);
            c = (int)r.x.ax;
            
        } while( r.h.cl != 0xFF && ( tmptr->tm_sec != target_second || (int)timenow.millitm < target_millisecond ) );

        if(r.h.cl != 0xFF)
            return -1;
        else
            return c;
    }

}                               /* timed_read_key */


int input_line( int buflen, char *buffer, int timeout, int *read_size )
{

struct timeb timenow;
struct tm *tmptr;
long c;
int row, col, target_second, target_millisecond;
char clo, chi;

    get_cursor_position( &row, &col );
    head_col = tail_col = col;
    init_char_pos = curr_char_pos = *read_size;
    ptr1 = ptr2 = end_ptr;

    if ( timeout != 0 ) {
        ftime( &timenow );
        tmptr = gmtime( &timenow.time );
        target_second = ( tmptr->tm_sec + timeout ) % 60;
        target_millisecond = timenow.millitm;
    }

    renablecursor();

    for ( ;; )
    {

        /* Read a single keystroke */

        do {
         
            if(timeout == 0)
                c = rgetchar();
            else {
                c = timed_read_key( target_second, target_millisecond );
                if ( c == 0 ) {
                    rdisablecursor();
                    return -1;
                }
            }
            
        } while(c == 0);

        clo = c;
        /* chi = (c >> 8) ;*/
        
        switch(clo) {

            case '\b':
            case ((char)127):

                /* Delete key action */
                if(*read_size == 0) {
                    
                    /* Ring bell if line is empty */
                    rbell();
                    
                } else {

                    /* Decrement read count */
                    --*read_size;
                    --curr_char_pos;

                    /* Erase last character typed */
                    get_cursor_position(&row, &col);
                    move_cursor(row, --col);
                    display_char((int)' ');
                    move_cursor(row, col);
                }
                break;
                
            case '\r':
            case '\n':
                rdisablecursor();
                scroll_line();
                return ( unsigned char ) clo;

            default:
                
                /* Normal key action */
                if(*read_size == ( buflen - 1 )) {

                    /* Ring bell if buffer is full */
                    rbell();
                    
                } else if ( clo < ' ' || clo >= 127 ) {
                    
                    rbell();

                } else {

                    /* Put key in buffer and display it */
                    buffer[curr_char_pos++] = clo;
                    if(*read_size < curr_char_pos)
                        *read_size = curr_char_pos;
                  
                    display_char((int)clo);

                }

        }

    }

} 


int input_character( int timeout )
{
struct timeb timenow;
struct tm *tmptr;
int target_second, target_millisecond;
int c;

    renablecursor();
    
    if(timeout != 0) {
        ftime(&timenow);
        tmptr = gmtime(&timenow.time);
        target_second = (tmptr->tm_sec + timeout) % 60;
        target_millisecond = timenow.millitm;
    }

    if(timeout == 0)
        c = inkey();
    else
        c = timed_read_key(target_second, target_millisecond);

    rdisablecursor();   

    return (char)c;
}                               /* input_character */


void scroll_line( void )
{
short left, top, right, bottom;
int row, col;

    get_cursor_position( &row, &col );
    _gettextwindow( &top, &left, &bottom, &right );
    
    if(row == bottom) {
        _settextwindow( status_size + 1, left, bottom, right );
        _scrolltextwindow( _GSCROLLUP );
        _settextwindow( top, left, bottom, right );
    } else
        row++;

    move_cursor( row, left );
}


/*
 * set_colours
 *
 * Sets the screen foreground and background colours.
 *
 */

void set_colours( zword_t foreground, zword_t background )
{
int fg, bg;
int colour_map[] = { BLACK, WHITE, WHITE, WHITE, WHITE, WHITE, WHITE, WHITE };

    if(foreground)
        fg = ( foreground == 1 ) ? WHITE : colour_map[foreground - 2];
    else
        fg = current_fg;

    if(background)
        bg = ( foreground == 1 ) ? ( fg == BLACK ) ? WHITE : BLACK : colour_map[background - 2];
    else
        bg = current_bg;

    /* Set foreground and background colour */

   _settextcolor( ( short ) fg | fgBright );
   _setbkcolor( ( long ) bg );

   /* Save new foreground and background colours for restoring colour */

   current_fg = ( int ) fg;
   current_bg = ( int ) bg;;

}  


/*
 * codes_to_text
 *
 * Translate Z-code characters to machine specific characters. These characters
 * include line drawing characters and international characters.
 *
 * The routine takes one of the Z-code characters from the following table and
 * writes the machine specific text replacement. The target replacement buffer
 * is defined by MAX_TEXT_SIZE in ztypes.h. The replacement text should be in a
 * normal C, zero terminated, string.
 *
 * Return 0 if a translation was available, otherwise 1.
 *
 *
 */

int codes_to_text( int c, char *s )
{
    /* Luckily, DEC properly implemented Latin1, unlike
     * IBM on their PC.  We can just use the zscii2latin1
     * table as is.
     */
    if(c > (char)154 || c < 224) {
        s[0] = zscii2latin1[c-(char)155];
        s[1] = '\0';
        return 0;
    } else
        return 1;
}                               /* codes_to_text */
